/***************************************************************************
                          MovingMesh.h  -  description
                             -------------------
    begin                : Wed Jun 20 2001
    copyright            : (C) 2001 by Robert Lie
    email                : rli@math.hkbu.edu.hk
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#ifndef _MovingMesh_H
#define _MovingMesh_H

#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <list>
#include <cmath>

#include <lac/precondition.h>
#include <lac/solver_control.h>
#include <lac/solver_bicgstab.h>
#include <lac/vector_memory.h>
#include <lac/sparse_ilu.h>

using namespace std;
#include <AFEPack/AMGSolver.h>



class LRTriangulation{
public:
	struct LRNode {
		int index;
		double coordinate[2];
		short boundary_mark;
	};
	struct LRFace {
		int node[2];
		int neighbour_element[2];
		short boundary_mark;
	};
	struct LRElement {
		int node[3];
		int face[3];
		int neighbour_element[3];
	};
public:
	string mesh_data_file;
	SparsityPattern spM;
	SparsityPattern spN;
	SparsityPattern spM_n;
	SparsityPattern spM_v;
	unsigned int n_max_coupling_node;
	unsigned int n_node;
	unsigned int n_face;
	unsigned int n_element;
	unsigned int n_interior_node;
	unsigned int n_boundary_node;
  unsigned int n_new_node;
	std::vector<int> interior_node_index;
	std::vector<int> boundary_node_index;
  std::vector<int> index;
	std::vector<LRNode> node;
	std::vector<LRFace> face;
	std::vector<LRElement> element;
public:
	LRTriangulation();
	virtual ~LRTriangulation();
	void ReadEasymesh(const string filename);
	
	std::vector<LRNode>::iterator begin_node() {return node.begin();}
	std::vector<LRNode>::iterator end_node() {return node.end();}
	
	std::vector<LRFace>::iterator begin_face() {return face.begin();}
	std::vector<LRFace>::iterator end_face() {return face.end();}
	
	std::vector<LRElement>::iterator begin_element() {return element.begin();}
	std::vector<LRElement>::iterator end_element() {return element.end();}
	
	LRNode& node_of_face(const LRFace& f, unsigned int i) 
		{assert(i < 2); return node[f.node[i]];}
	LRNode& node_of_element(const LRElement& e, unsigned int i) 
		{assert(i < 3); return node[e.node[i]];}
	LRFace& face_of_element(const LRElement& e, unsigned int i) 
		{assert(i < 3); return face[e.face[i]];}
	LRElement * neighbour_element_of_face(const LRFace& f, unsigned i) {
		assert(i < 2); 
		if (f.neighbour_element[i] == -1)
			return 0;
		else
			return &element[f.neighbour_element[i]];
	}
	LRElement * neighbour_element_of_element(const LRElement& e, unsigned i) {
		assert(i < 3); 
		if (e.neighbour_element[i] == -1)
			return 0;
		else
			return &element[e.neighbour_element[i]];
	}
	
	LRNode& interior_node(unsigned int i) 
		{assert(i < n_interior_node); return node[interior_node_index[i]];}
	LRNode& boundary_node(unsigned int i) 
		{assert(i < n_boundary_node); return node[boundary_node_index[i]];}
	
private:
	void ParseMesh();
};	
	
	
	


class LRMovingMesh : public LRTriangulation
{	
public:
	struct LRPoint {
		double coordinate[2];
	};
	struct LRVertex {
		int index;
		double coordinate[2];
		short boundary_mark;
	};
	struct LREdge {
		int index;
		int vertex[2];
		short boundary_mark;
	};
	struct LRDomain {
		unsigned int n_vertex;
		unsigned int n_edge;
		std::vector<LRVertex> physical_domain_vertex;
		std::vector<LRVertex> logical_domain_vertex;
		std::vector<LREdge> edge;
	};
public:
	LRDomain domain;
	double move_step_length;
	std::vector<LRPoint> logical_node;

	std::vector<LRPoint> move_direction;
	std::vector<LRPoint> logical_move_direction;
	std::vector<double> monitor_function;
public:
	LRMovingMesh();
	virtual ~LRMovingMesh();
	int MoveMesh();
	void OutputPhysicalMesh(const string filename);
	void OutputLogicalMesh(const string filename);
	virtual void MonitorFunction();
	virtual void SmoothMonitor(unsigned int s = 1);
	virtual void UpdateSolution();
	void ReadDomain(const string filename);

private:
	void GetLogicalMesh();
	void GetMoveDirection();
	void GetMoveStepLength();
	void ReadDummy(ifstream& is);
};


#endif
